!=========================================================================
! Copyright (C) GemTalk Systems 1986-2019.  All Rights Reserved.
!
! $Id: reposit2.gs 40520 2016-10-20 17:20:32Z bretlb $
! Superclass Hierarchy:
!   Repository, Collection, Object.
!
!=========================================================================

category: 'Backup and Restore'
method: Repository
lastFailoverTime

"If system is in restore from transaction logs, returns the
 last failOver timestamp detected while reading transaction logs,
 or nil if no failOver timestamp has been processed since restore started.
 If system is not in restore, returns the time from the most 
 recent suspendCommitsForFailover, or nil if no suspendCommitsForFailover
 has been executed.

 Result is a DateAndTime or nil."

 | info tm suspendTime |
 ((info := self restoreStatusInfo) at: 2) == 2 ifTrue:[
   (tm := info at: 13) == 0 ifTrue:[ ^ nil ]. 
   ^ DateAndTime posixSeconds: tm offset: Duration zero
 ].
 (suspendTime := info at: 14) == 0 ifTrue:[ ^ nil ].
 ^ DateAndTime posixSeconds: suspendTime offset: Duration zero
% 

category: 'Private'
method: Repository
_doScan: kind fast: isFast with: inputArgument

" This method calls does the appropriate tests, and calls the the _scan primitive 
  which does the bulk of the work.

   scanKind  Function     
   =======================
     0     allInstances  
     1     allReferences
     2     allReferencesToInstancesOfClasses
     3     allObjectsInObjectSecurityPolicies
     4     allObjectsLargerThan
"
| maxThreads percentCpu argClass result inputArg |
inputArg := (inputArgument _isArray and: [inputArgument size > 1])
                ifTrue:[ Array withAll:(IdentitySet withAll: inputArgument)]
                ifFalse:[ inputArgument ].
System needsCommit ifTrue: [ System _error: #rtErrAbortWouldLoseData ] .

isFast ifTrue: [ maxThreads := self _aggressiveMaxThreadCount.
                 percentCpu := 95 ]
       ifFalse: [ maxThreads := 2. percentCpu := 90 ].
result := self _scanWithMaxThreads: maxThreads waitForLock: 60 pageBufSize: 8
               percentCpuActiveLimit: percentCpu scanKind: kind with: inputArg .
argClass := inputArg class.
((argClass == Array) or: [argClass == GsBitmap]) ifFalse: [
    result := (result at: 1) at: 2. ] .
^result
%
category: 'Repository Analysis'
method: Repository
loadGcCandidatesFromFile: aString intoHiddenSet: hiddenSetId

"Given a file which is binary list of dead objects (produced by one of the 
 #findDisconnectedObjects: methods), load the objects into the hidden set 
 with the given ID.  NOTE: in 3.4 the hiddenSetId changed from a SmallInteger to
 one of the symbols in GsBitmap>>hiddenSetSpecifiers.

 Only objects which exist and are not in the stone's dead object
 set are added to the hidden set.

 Used by GsObjectInventory to profile garbage in the repository.

 Returns an Array containing 3 SmallIntegers:
  1 - Total number of objects in the file.
  2 - Number of valid objects loaded into the hidden set.
  3 - Number of invalid objects not loaded.

 Requires the GarbageCollection privilege.

 Signals an error if hiddenSetId specifies a hidden set only writable
 by SystemUser and the session is not logged in as SystemUser.
"

| gsBitmap numInvalid result bmSize |

gsBitmap := GsBitmap newForHiddenSet: hiddenSetId.
numInvalid := gsBitmap readFromFile: aString.
bmSize := gsBitmap size.
result := Array with: (bmSize + numInvalid)
                with: bmSize 
                with: numInvalid.
^result
%

category: 'Deprecated (Single Ref Path)'
method: Repository
findReferencePathToObject: anObject

"Finds a reference path to an object.  Returns an instance of GsSingleRefPathResult, 
 for which each element represents a connection leading between the search 
 objects and the reference in the repository.  
 The first element will usually be one of the objects in the limit set, and the last 
 element will be the argument. 

 The searchObj must be a committed non-special object. This method is not to be used 
 for classes or metaclasses. Objects in the stone's list of dead objects are excluded from 
 the scan.  It is an error if the object is in the stone's dead objects set. 

 Only a single reference path to each object is returned. It is possible for any object 
 to be connected to the repository through multiple reference paths.

 It is possible for an object to be disonnected from the repository and therefore have no
 reference path. In this case, the result GsSingleRefPathResult isDead will be true. 

 This method aborts the current transaction; if an abort would cause unsaved
 changes to be lost, it signals an error, #rtErrAbortWouldLoseData.

 A reference path scan for an object is terminated and considered completed if any class  
 object or GsObjectSecurityPolicy is found in the reference path.  Normally all classes are 
 connected to the repository and will be included in the limit set.  However, for garbage 
 collection purposes, disconnected classes are considered live objects if any other object 
 references them.  In other words, a class is a live object if any live instances of that 
 class exist in the repository.
"

self deprecated: 'Repository>>findReferencePathToObject: deprecated v3.4. ',
 'Use (GsSingleRefPathFinder newForSearchObjects: {anObject}) runScan buildResultObjects first'.
	
System needsCommit ifTrue: [ System _error: #rtErrAbortWouldLoseData ] .

^ (GsSingleRefPathFinder newForSearchObjects: { anObject }) runScan
      buildResultObjects first
%

category: 'Deprecated (Single Ref Path)'
method: Repository
findReferencePathToObjects: anArray findAllRefs: findAllRefs printToLog: printToLog

"Finds a reference path to an array of objects.  Returns an array of GsSingleRefPathResult
 showing how each object is connected to the repository by reference.  

 Note that arguments to this method other than <anArray> are not used in v3.4.

 See the comments in findReferencePathToObject: for more information.

 This method aborts the current transaction; if an abort would cause unsaved
 changes to be lost, it signals an error, #rtErrAbortWouldLoseData."

self deprecated: 'Repository>>findReferencePathToObjects:findAllRefs:printToLog: deprecated v3.4. ',
 'Use (GsSingleRefPathFinder newForSearchObjects: anArray) runScan buildResultObjects'.
		 
System needsCommit ifTrue: [ self _error: #rtErrAbortWouldLoseData ] .

^ (GsSingleRefPathFinder newForSearchObjects: anArray) runScan
      buildResultObjects 
%

category: 'Deprecated (Single Ref Path)'
method: Repository
findReferencePathToObjs: anArray limitObjArray: limitArray findAllRefs: findAllRefs printToLog: printToLog

"Finds a reference path to an array of objects.  Returns an array of GsSingleRefPathResult
 showing how each object is connected to the repository by reference.

 Note that arguments to this method other than <anArray> are not used in v3.4.

 See the comments in findReferencePathToObject: for more information.

 This method aborts the current transaction; if an abort would cause unsaved
 changes to be lost, it signals an error, #rtErrAbortWouldLoseData."

self deprecated: 'Repository>>findReferencePathToObjs:limitObjArray:findAllRefs:printToLog: deprecated v3.4. ',
 'Use (GsSingleRefPathFinder newForSearchObjects: anArray) runScan buildResultObjects'.

System needsCommit ifTrue: [ self _error: #rtErrAbortWouldLoseData ] .

^ (GsSingleRefPathFinder newForSearchObjects: anArray) runScan
      buildResultObjects 
%

category 'Private'
method: Repository
_getNumThreads
  ^ SessionTemps current at: #GsOverrideNumThreads 
      ifAbsent: [ (SystemRepository numberOfExtents) * 2 ]
%

category: 'Private'
method
_buildListInstResult: allInstResult

 | gsBitmap arrSize element resultArr |

 gsBitmap := GsBitmap newForHiddenSet: #ListInstances.
 gsBitmap removeAll.

 arrSize := allInstResult size.
 arrSize == 1 ifTrue: [
   element := allInstResult at: 1.
   gsBitmap addAll: (element at: 2).
   ^ (element at: 2) size
 ].

 resultArr := Array new.
 1 to: arrSize do: [ :i |
   element := allInstResult at: i.
   resultArr add: (element at: 1).
   resultArr add: ((element at: 2) size).
   gsBitmap addAll: (element at: 2).
 ].
 ^ resultArr.
%

category: 'Private'
method
_buildAllRefsResult: allRefsResult

 | resultArr |
 resultArr := Array new.
 1 to: (allRefsResult size) do: [ :i |  | element |
   element := allRefsResult at: i.
   resultArr add: ((element at: 2) asArray).
 ].
 ^ resultArr.
%

category: 'Private'
method
_buildAllRefsResult: allRefsResult withLimit: aSmallInt

 | resultArr |
 resultArr := Array new.
 1 to: (allRefsResult size) do: [ :i |  | element |
   element := allRefsResult at: i.
   resultArr add: ((element at: 2) _enumerateWithLimit: aSmallInt startingAfter: 0)
 ].
 ^ resultArr.
%

